from functools import lru_cache

import sass
from django.conf import settings
from django.contrib.staticfiles.finders import get_finders
from pipeline.compilers import CompilerBase

from .sass_functions.svg_to_data_uri import svg_to_data_uri


def svg_to_data_uri_(file_):
    return svg_to_data_uri(file_, get_include_paths())


OUTPUT_STYLE = getattr(settings, 'LIBSASS_OUTPUT_STYLE', 'nested')
SOURCE_COMMENTS = getattr(settings, 'LIBSASS_SOURCE_COMMENTS', settings.DEBUG)
CUSTOM_FUNCTIONS = getattr(settings, 'LIBSASS_CUSTOM_FUNCTIONS', {
    'svg_data_uri': svg_to_data_uri_,
})


@lru_cache()
def get_include_paths():
    """
    Generate a list of include paths that libsass should use to find files
    mentioned in @import lines.
    """
    include_paths = []

    # Look for staticfile finders that define 'storages'
    for finder in get_finders():
        try:
            storages = finder.storages
        except AttributeError:
            continue

        for storage in storages.values():
            try:
                include_paths.append(storage.path('.'))
            except NotImplementedError:  # pragma: no cover
                # storages that do not implement 'path' do not store files locally,
                # and thus cannot provide an include path
                pass
    return include_paths


def compile(**kwargs):
    """Perform sass.compile, but with the appropriate include_paths for Django added"""
    INCLUDE_PATHS = get_include_paths()

    kwargs = kwargs.copy()
    kwargs['include_paths'] = (kwargs.get('include_paths') or []) + INCLUDE_PATHS

    custom_functions = CUSTOM_FUNCTIONS.copy()
    custom_functions.update(kwargs.get('custom_functions', {}))
    kwargs['custom_functions'] = custom_functions

    return sass.compile(**kwargs)


class SassCompiler(CompilerBase):
    output_extension = 'css'

    def match_file(self, filename):
        return filename.endswith('.scss')

    def compile_file(self, infile, outfile, outdated=False, force=False):
        # if not outdated and not force:
        #     return
        with open(outfile, 'w') as f:
            f.write(
                compile(
                    filename=infile,
                    output_style=OUTPUT_STYLE,
                    source_comments=SOURCE_COMMENTS,
                )
            )
        return outfile



from unittest.mock import patch

from django.test import TestCase

from ..raven import capture_exception


class RavenTestCase(TestCase):

    @capture_exception
    def func_success(self, arg1, kwarg2):
        return 'Good {} {}'.format(arg1, kwarg2)

    @capture_exception
    def func_fail(self):
        raise ValueError('Fail')

    def test_success(self):
        with patch('utils.raven.client.captureException') as mock:
            self.assertEqual(self.func_success('arg1', kwarg2='kwarg2'), 'Good arg1 kwarg2')
        self.assertFalse(mock.called)

    def test_fail(self):
        with patch('utils.raven.client.captureException') as mock:
            with self.assertRaises(ValueError):
                self.func_fail()
        self.assertTrue(mock.called)



import re

import yaml

FM_BOUNDARY = re.compile(r'^---$', re.MULTILINE)


def split(content):
    try:
        _, fm, body = FM_BOUNDARY.split(content, 2)
    except ValueError:
        raise ValueError('Invalid front matter content')
    try:
        fm_data = yaml.safe_load(fm)
    except yaml.scanner.ScannerError:
        raise ValueError('Invalid yaml')
    return [
        fm_data,
        body[1:],
    ]



from django.test import TestCase

from ..frontmatter import split


template = '''
---
key: value
foo:
  - bar
  - baz
---
Body content
'''

template_invalid = '''
---
- this
is
- invalid
---
'''

template_missing_frontmatter = '''
Nothing here
'''


class FrontmatterTestCase(TestCase):

    def test_success(self):
        data, body = split(template)
        self.assertEquals(data, {
            'key': 'value',
            'foo': ['bar', 'baz'],
        })
        self.assertEquals(body, 'Body content\n')

    def test_invalid_yaml(self):
        with self.assertRaisesRegex(ValueError, 'Invalid yaml'):
            split(template_invalid)

    def test_missing_frontmatter(self):
        with self.assertRaisesRegex(ValueError, 'Invalid front matter content'):
            split(template_missing_frontmatter)



import base64

from django.core.signing import Signer, TimestampSigner


def fix_missing_padding(s):
    # For some reason, some browsers go to the url without the padding =
    missing_padding = len(s) % 4
    if missing_padding != 0:
        s += '=' * (4 - missing_padding)
    return s


class UrlSafeMixin(object):
    def sign(self, value):
        value = super().sign(str(value)).encode('utf-8')
        return base64.urlsafe_b64encode(value).decode('utf-8')

    def unsign(self, value, *args, **kwargs):
        value = base64.urlsafe_b64decode(fix_missing_padding(value)).decode('utf-8')
        return super().unsign(value, *args, **kwargs)


class UrlSafeTimestampSigner(UrlSafeMixin, TimestampSigner):
    pass


class UrlSafeSigner(UrlSafeMixin, Signer):
    pass


default_signer = UrlSafeTimestampSigner()
sign = default_signer.sign
unsign = default_signer.unsign



from django import template

from utils.urls import absolute_reverse

register = template.Library()


@register.simple_tag()
def absolute_url(viewname, *args, **kwargs):
    return absolute_reverse(viewname, args=args, kwargs=kwargs)



import mimetypes
from io import StringIO

import boto3
from django.conf import settings
from django.contrib.staticfiles.storage import (
    ManifestFilesMixin, StaticFilesStorage)
from django.core.cache import cache
from django.core.files.base import ContentFile
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from pipeline.storage import PipelineMixin


class PipelineManifestStorage(PipelineMixin, ManifestFilesMixin, StaticFilesStorage):
    pass


class S3HeadCacheMixin:
    def get_cache_key(self, name):
        return 'storages_meta__{}'.format(name)

    def get_head(self, name, use_cache=True):
        if use_cache:
            data = cache.get(self.get_cache_key(name))
            if data:
                return data
        data = self.s3client.head_object(Bucket=self.bucket, Key=name)
        cache.set(self.get_cache_key(name), data)
        return data

    def reset_head(self, name):
        cache.delete(self.get_cache_key(name))


@deconstructible
class S3MediaStorage(Storage, S3HeadCacheMixin):
    def __init__(self):
        self.s3client = boto3.client('s3', region_name=settings.AWS_REGION)
        self.bucket = settings.MEDIA_BUCKET

    def delete(self, name):
        self.s3client.delete_object(Bucket=self.bucket, Key=name)

    def exists(self, name):
        try:
            self.get_head(name, use_cache=False)
            return True
        except Exception:
            return False

    def size(self, name):
        return self.get_head(name)['ContentLength']

    def url(self, name):
        return 'https://s3-{aws_region}.amazonaws.com/{bucket}/{key}'.format(
            aws_region=settings.AWS_REGION,
            bucket=self.bucket,
            key=name,
        )

    def get_modified_time(self, name):
        return self.get_head(name)['LastModified']

    def _open(self, name, mode='rb'):
        response = self.s3client.get_object(Bucket=self.bucket, Key=name)
        return ContentFile(response['Body'].read())

    def _save(self, name, content):
        if isinstance(content.file, StringIO):
            content = ContentFile(content.read().encode('utf-8'), content.name)
        self.s3client.put_object(
            Bucket=self.bucket,
            Key=name,
            Body=content,
            ACL='public-read',
            ContentType=mimetypes.guess_type(name)[0] or 'binary/octet-stream'
        )
        self.reset_head(name)
        return name



from django.template import Context, Template
from django.test import TestCase

template = '{% load jsonify %}{{ content | jsonify:2 }}'

template_html = '''[
  "foo",
  "bar",
  "ba<"
]'''


class MarkdownTestCase(TestCase):
    def test_markdown(self):
        data = {
            'content': [
                'foo',
                'bar',
                'ba<'
            ],
        }
        html = Template(template).render(Context(data))
        self.assertEqual(html, template_html)



import json

from django.contrib.postgres.forms.jsonb import InvalidJSONInput, JSONField
from django.forms import ValidationError


class JSONPrettyField(JSONField):
    def __init__(self, *args, **kwargs):
        self.__indent = kwargs.pop('indent', 2)
        self.__dict_only = kwargs.pop('dict_only', False)
        self.__list_only = kwargs.pop('list_only', False)
        if self.__dict_only and self.__list_only:
            raise ValueError('Only one of dict_only or list_only can be True')
        super().__init__(*args, **kwargs)

    def prepare_value(self, value):
        if isinstance(value, InvalidJSONInput):
            return value
        return json.dumps(value, indent=self.__indent, sort_keys=True, ensure_ascii=False)

    def validate(self, value):
        if self.__dict_only and not isinstance(value, dict):
            raise ValidationError('Value is not of type dict')
        if self.__list_only and not isinstance(value, list):
            raise ValidationError('Value is not of type list')
        return value


class YAMLPrettyField(JSONPrettyField):
    def __init__(self, *args, **kwargs):
        self.__indent = kwargs.pop('indent', 2)
        self.__default_flow_style = kwargs.pop('default_flow_style', False)
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        import yaml
        try:
            return yaml.safe_load(value)
        except yaml.YAMLError as e:
            raise ValidationError(e)

    def prepare_value(self, value):
        import yaml
        try:
            return yaml.safe_dump(
                value,
                indent=self.__indent,
                default_flow_style=self.__default_flow_style,
                allow_unicode=True,
            )
        except Exception:
            return value



from django.template import Context, Template
from django.test import TestCase

template_filter = '{% load md %}{{ content | markdown }}'
template_tag = '{% load md %}{% markdown %}{{ content }}{% endmarkdown %}'
template_tag_as = '{% load md %}{% markdown as my_variable %}{{ content }}{% endmarkdown %}'

content = '''
# My template

Some test
'''

content_html = '''
<h1>My template</h1>
<p>Some test</p>
'''.strip()


class MarkdownTestCase(TestCase):
    def test_filter(self):
        data = {
            'content': content
        }
        html = Template(template_filter).render(Context(data))
        self.assertEqual(html, content_html)

    def test_template_tag(self):
        data = {
            'content': content
        }
        html = Template(template_tag).render(Context(data))
        self.assertEqual(html, content_html)

    def test_template_tag_as(self):
        data = {
            'content': content
        }
        context = Context(data)
        html = Template(template_tag_as).render(context)
        self.assertEqual(html, '')
        self.assertIn('my_variable', context)
        self.assertEqual(context['my_variable'], content_html)



import os
from io import StringIO

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase


class PipelineTestCase(TestCase):
    def setUp(self):
        file_path = os.path.join(settings.STATIC_ROOT, 'staticfiles.json')
        if os.path.isfile(file_path):
            os.remove(file_path)

    def test_success(self):
        call_command('collectstatic', '--noinput', stdout=StringIO())
        call_command('clean_staticfilesjson', stdout=StringIO())

    def test_missing_staticfilesjson(self):
        with self.assertRaises(CommandError):
            call_command('clean_staticfilesjson', stdout=StringIO())



from functools import wraps
from itertools import chain

from django.core.cache import cache


def create_cache_key(func, args, kwargs):
    return '-'.join(
        map(
            str,
            chain(
                [func.__qualname__],
                args,
                map(':'.join, kwargs.items()),
            )
        )
    )


class cache_function:
    def __init__(self, timeout, key=None):
        self.timeout = timeout
        self.key = key

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = self.key
            if key is None:
                key = create_cache_key(func, args, kwargs)
            elif callable(key):
                key = create_cache_key(func, [key(*args, **kwargs)], {})
            data = cache.get(key)
            if data:
                return data
            data = func(*args, **kwargs)
            cache.set(key, data, self.timeout)
            return data
        return wrapper



from unittest.mock import patch

from django.test import TestCase

from ..mixins import SignedPkMixin
from ..signing import UrlSafeSigner, sign, unsign


class ExampleClass(SignedPkMixin):
    def __init__(self, **kwargs):
        self.kwargs = kwargs


class SignedPkMixinTestCase(TestCase):
    def test_pk(self):
        obj = ExampleClass(pk='123')
        with patch('utils.mixins.unsign') as mock:
            self.assertEqual(obj.get_pk(), '123')
        self.assertFalse(mock.called)
        self.assertEqual(unsign(obj.get_pk_signed()), '123')

    def test_pk_signed(self):
        obj = ExampleClass(pk_signed=sign('123'))
        self.assertEqual(obj.get_pk(), '123')
        with patch('utils.mixins.unsign') as mock:
            self.assertEqual(unsign(obj.get_pk_signed()), '123')
        self.assertFalse(mock.called)

    def test_bad_base64(self):
        signer = UrlSafeSigner()
        pk_signed = str(signer.sign(123))
        self.assertIn('=', pk_signed)
        pk_signed = pk_signed.replace('=', '')
        self.assertEqual(signer.unsign(pk_signed), '123')



import yaml

from django import forms
from django.test import TestCase

from ..fields import YAMLPrettyField


class TestForm(forms.Form):
    data_field = YAMLPrettyField(indent=4)


class TestDictForm(forms.Form):
    data_field = YAMLPrettyField(indent=4, dict_only=True)


class TestListForm(forms.Form):
    data_field = YAMLPrettyField(indent=4, list_only=True)


class YAMLPrettyFieldTestCase(TestCase):
    data_dict = 'id: 1\ntest: 2'
    data_list = '- id: 1\n- id: 2'

    def test_indentation(self):
        data = {'id': 1, 'name': {'first': 'Foo', 'last': 'Bar'}}
        form_data = {'data_field': data}
        form = TestForm(initial=form_data)
        self.assertIn(yaml.safe_dump(data, indent=4, default_flow_style=False), form.as_p())

    def test_dict_only_ok(self):
        form_data = {'data_field': self.data_dict}
        form = TestDictForm(form_data)
        self.assertTrue(form.is_valid())

    def test_dict_only_fail(self):
        form_data = {'data_field': self.data_list}
        form = TestDictForm(form_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['data_field'], ['Value is not of type dict'])

    def test_list_only_ok(self):
        form_data = {'data_field': self.data_list}
        form = TestListForm(form_data)
        self.assertTrue(form.is_valid())

    def test_list_only_fail(self):
        form_data = {'data_field': self.data_dict}
        form = TestListForm(form_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['data_field'], ['Value is not of type list'])

    def test_prepare_value_fail(self):
        form_data = {'data_field': 'a: b, c: d'}
        form = TestForm(form_data)
        self.assertIn('a: b, c: d', form.as_p())



import logging

import requests

from django.core.cache import cache


def shorten(url):
    KEY = 'shorturl:{}'.format(url)
    logger = logging.getLogger(__name__)
    logger.info('Shortening url {}'.format(url))
    shorturl = cache.get(KEY)
    if shorturl:
        logger.info('Returning cached shorturl {}'.format(shorturl))
        return shorturl
    data = requests.get('https://is.gd/create.php', params={'format': 'json', 'url': url}).json()
    if 'shorturl' in data:
        shorturl = data['shorturl']
        logger.info('Shorturl: {}'.format(shorturl))
        # Store shorturl for 60 days
        cache.set(KEY, shorturl, 60 * 24 * 3600)
    else:
        logger.warning(data.get('errormessage', 'shorturl errormessage was not found'))
        # Store original url for 1 hour
        cache.set(KEY, url, 3600)
        return url
    return shorturl



from itertools import chain, count

from django.test import TestCase

from ..decorators import cache_function

i = count()


def func_no_args_no_cache():
    return next(i)


@cache_function(3600)
def func_no_args():
    return func_no_args_no_cache()


@cache_function(3600)
def func_args(*args):
    return ', '.join(chain([str(func_no_args_no_cache())], args[::-1]))


@cache_function(3600)
def func_kwargs(a, b):
    return ''.join([
        'a=',
        a,
        ' b=',
        b,
        str(func_no_args_no_cache()),
    ])


@cache_function(3600, key=lambda *args: sum(args))
def func_key(a, b):
    return func_no_args_no_cache()


class CacheFunctionTestCase(TestCase):
    def test_func_no_args(self):
        self.assertNotEqual(func_no_args_no_cache(), func_no_args_no_cache())
        self.assertEqual(func_no_args(), func_no_args())

    def test_func_args(self):
        self.assertEqual(func_args('a', 'b'), func_args('a', 'b'))
        self.assertNotEqual(func_args('a', 'b'), func_args('b', 'a'))

    def test_func_kwargs(self):
        self.assertEqual(func_kwargs(a='a', b='b'), func_kwargs(a='a', b='b'))
        self.assertNotEqual(func_kwargs('a', 'b'), func_kwargs(a='a', b='b'))

    def test_func_key(self):
        self.assertEqual(func_key(1, 2), func_key(1, 2))
        self.assertEqual(func_key(2, 1), func_key(1, 2))



# !/usr/bin/env python

from itertools import chain
from os import path

from setuptools import find_packages, setup

extras_require = {
    'markdown': [
        'Markdown>=2.6,<3'
    ],
    'pipeline': [
        'django-pipeline>=1.6,<1.7',
        'libsass>=0.11,<0.13',
        'csscompressor>=0.9,<1.0',
    ],
    'raven': [
        'raven>=6.0,<6.1'
    ],
    's3': [
        'boto3>=1.4,<1.5',
    ],
    'frontmatter': [
        'PyYAML>=3,<4',
    ],
    'json': [
        'psycopg2-binary>=2.7',
    ],
}

extras_require['all'] = list(chain(extras_require.values()))

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bulv1ne-django-utils',
    packages=find_packages(exclude=['tests']),
    version='0.2.5',
    description='Django utils for all kinds of situations',
    long_description=long_description,
    author='Niels Lemmens',
    license='MIT',
    author_email='draso.odin@gmail.com',
    url='https://github.com/bulv1ne/django-utils',
    keywords=['django', 'utils'],
    install_requires=[
        'django>=1.10',
    ],
    extras_require=extras_require,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
)



from django.core.exceptions import ValidationError
from django.core.validators import EmailValidator
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_text


@deconstructible
class EmailListValidator(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, value):
        email_list = force_text(value).split()
        email_validator = EmailValidator(*self.args, **self.kwargs)
        for email in email_list:
            try:
                email_validator(email)
            except ValidationError as e:
                raise ValidationError(
                    '"{}" is not a valid email'.format(email),
                    code=e.code,
                )



import json

from django import forms
from django.test import TestCase
from django.utils.html import escape

from ..fields import JSONPrettyField


class TestForm(forms.Form):
    data_field = JSONPrettyField(indent=4)


class TestDictForm(forms.Form):
    data_field = JSONPrettyField(indent=4, dict_only=True)


class TestListForm(forms.Form):
    data_field = JSONPrettyField(indent=4, list_only=True)


class JSONPrettyFieldTestCase(TestCase):
    def test_indentation(self):
        data = [{'id': 1}, {'id': 2}]
        form_data = {'data_field': json.dumps(data)}
        form = TestForm(form_data)
        self.assertIn(escape(json.dumps(data, indent=4)), form.as_p())

    def test_dict_only_ok(self):
        data = {'id': 1}
        form_data = {'data_field': json.dumps(data)}
        form = TestDictForm(form_data)
        self.assertTrue(form.is_valid())

    def test_dict_only_fail(self):
        data = [{'id': 1}, {'id': 2}]
        form_data = {'data_field': json.dumps(data)}
        form = TestDictForm(form_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['data_field'], ['Value is not of type dict'])

    def test_list_only_ok(self):
        data = [{'id': 1}, {'id': 2}]
        form_data = {'data_field': json.dumps(data)}
        form = TestListForm(form_data)
        self.assertTrue(form.is_valid())

    def test_list_only_fail(self):
        data = {'id': 1}
        form_data = {'data_field': json.dumps(data)}
        form = TestListForm(form_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['data_field'], ['Value is not of type list'])

    def test_dict_and_list(self):
        with self.assertRaises(ValueError):
            JSONPrettyField(dict_only=True, list_only=True)

    def test_prepare_value_fail(self):
        form_data = {'data_field': 'invalid'}
        form = TestForm(form_data)
        self.assertIn('invalid', form.as_p())
